@@ -233,7 +233,8 @@ import { TextInput as TextInputPF4 } from "@patternfly/react-core/dist/esm/compo
 import { Popover } from "@patternfly/react-core/dist/esm/components/Popover/index.js";
 import { HelperText, HelperTextItem } from "@patternfly/react-core/dist/esm/components/HelperText/index.js";
 import { List, ListItem } from "@patternfly/react-core/dist/esm/components/List/index.js";
-import { ExclamationTriangleIcon, InfoIcon, HelpIcon } from "@patternfly/react-icons";
+import { ExclamationTriangleIcon, InfoIcon, HelpIcon, EyeIcon, EyeSlashIcon } from "@patternfly/react-icons";
+import { InputGroup } from "@patternfly/react-core/dist/esm/components/InputGroup/index.js";
 
 import { show_modal_dialog, apply_modal_dialog } from "cockpit-components-dialog.jsx";
 import { ListingTable } from "cockpit-components-table.jsx";
@@ -353,12 +354,10 @@ export const dialog_open = (def) => {
     };
 
     const props = () => {
-        const title = (def.Action && (def.Action.Danger || def.Action.DangerButton)
-            ? <><ExclamationTriangleIcon className="ct-icon-exclamation-triangle" /> {def.Title}</>
-            : def.Title);
         return {
             id: "dialog",
-            title,
+            title: def.Title,
+            titleIconVariant: (def.Action && (def.Action.Danger || def.Action.DangerButton)) ? "warning" : null,
             body: <Body body={def.Body}
                         teardown={def.Teardown}
                         fields={nested_fields}
@@ -401,6 +400,7 @@ export const dialog_open = (def) => {
                         }
                         errors = errs;
                         update();
+                        update_footer();
                         return Promise.reject();
                     });
         };
@@ -408,28 +408,32 @@ export const dialog_open = (def) => {
     }
 
     const footer_props = (running_title, running_promise) => {
-        let actions = [];
-        if (def.Action) {
-            actions = [
-                {
-                    caption: def.Action.Title,
-                    style: "primary",
-                    danger: def.Action.Danger || def.Action.DangerButton,
-                    disabled: running_promise != null,
-                    clicked: progress_callback => run_action(progress_callback, null),
-                }
-            ];
+        const actions = [];
 
-            if (def.Action.Variants)
+        function add_action(variant) {
+            actions.push({
+                caption: variant.Title,
+                style: actions.length == 0 ? "primary" : "secondary",
+                danger: def.Action.Danger || def.Action.DangerButton,
+                disabled: running_promise != null || (def.Action.disable_on_error &&
+                                                      errors && errors.toString() != "[object Object]"),
+                clicked: progress_callback => run_action(progress_callback, variant.tag),
+            });
+        }
+
+        if (def.Action) {
+            if (def.Action.Title) {
+                add_action({
+                    Title: def.Action.Title,
+                    tag: null,
+                });
+            }
+
+            if (def.Action.Variants) {
                 for (const v of def.Action.Variants) {
-                    actions.push({
-                        caption: v.Title,
-                        style: "secondary",
-                        danger: def.Action.Danger || def.Action.DangerButton,
-                        disabled: running_promise != null,
-                        clicked: progress_callback => run_action(progress_callback, v.tag),
-                    });
+                    add_action(v);
                 }
+            }
         }
 
         const extra = (
@@ -506,6 +510,14 @@ export const dialog_open = (def) => {
             update();
         },
 
+        get_options: (tag) => {
+            for (const f of fields) {
+                if (f.tag == tag) {
+                    return f.options;
+                }
+            }
+        },
+
         set_options: (tag, new_options) => {
             fields.forEach(f => {
                 if (f.tag == tag) {
@@ -563,6 +575,27 @@ export const TextInput = (tag, title, options) => {
     };
 };
 
+const PassInputElement = ({ tag, title, options, val, change, validated }) => {
+    const [show, setShow] = useState(false);
+
+    return (
+        <InputGroup>
+            <TextInputPF4 data-field={tag} data-field-type="text-input"
+                          validated={validated}
+                          disabled={options.disabled}
+                          aria-label={title}
+                          autoComplete={options.new_password ? "new-password" : null}
+                          type={show ? "text" : "password"}
+                          value={val}
+                          onChange={(_event, value) => change(value)} />
+            <Button variant="control"
+                    onClick={() => setShow(!show)}
+                    isDisabled={options.disabled}>
+                { show ? <EyeSlashIcon /> : <EyeIcon /> }
+            </Button>
+        </InputGroup>);
+};
+
 export const PassInput = (tag, title, options) => {
     return {
         tag,
@@ -571,12 +604,12 @@ export const PassInput = (tag, title, options) => {
         initial_value: options.value || "",
 
         render: (val, change, validated) =>
-            <TextInputPF4 data-field={tag} data-field-type="text-input"
-                   validated={validated}
-                   disabled={options.disabled}
-                   aria-label={title}
-                   type="password" value={val}
-                   onChange={(_event, value) => change(value)} />
+            <PassInputElement tag={tag}
+                              title={title}
+                              options={options}
+                              val={val}
+                              change={change}
+                              validated={validated} />
     };
 };
 
@@ -660,6 +693,28 @@ export const SelectOneRadio = (tag, title, options) => {
     };
 };
 
+export const SelectOneRadioVertical = (tag, title, options) => {
+    return {
+        tag,
+        title,
+        options,
+        initial_value: options.value || options.choices[0].value,
+        hasNoPaddingTop: true,
+
+        render: (val, change) => {
+            return (
+                <div data-field={tag} data-field-type="select-radio">
+                    { options.choices.map(c => (
+                        <Radio key={c.value} isChecked={val == c.value} data-data={c.value}
+                            id={tag + '.' + c.value}
+                            onChange={() => change(c.value)} label={c.title} />))
+                    }
+                </div>
+            );
+        }
+    };
+};
+
 export const SelectRow = (tag, headers, options) => {
     return {
         tag,
@@ -700,8 +755,9 @@ export const SelectSpaces = (tag, title, options) => {
         tag,
         title,
         options,
-        initial_value: [],
+        initial_value: options.value || [],
         hasNoPaddingTop: options.spaces.length == 0,
+
         render: (val, change) => {
             if (options.spaces.length === 0)
                 return <span className="text-danger">{options.empty_warning}</span>;
@@ -715,20 +771,15 @@ export const SelectSpaces = (tag, title, options) => {
                         const desc = block === spc.desc ? "" : spc.desc;
 
                         const on_change = (_event, checked) => {
+                            // Be careful to keep "val" in the same order as "options.spaces".
                             if (checked && !selected)
-                                change(val.concat(spc));
+                                change(options.spaces.filter(v => val.indexOf(v) >= 0 || v == spc));
                             else if (!checked && selected)
                                 change(val.filter(v => (v != spc)));
                         };
 
-                        return (
-                            <DataListItem key={spc.block ? spc.block.Device : spc.desc}>
-                                <DataListItemRow>
-                                    <DataListCheck id={(spc.block ? spc.block.Device : spc.desc) + "-row-checkbox"}
-                                                   isChecked={selected} onChange={on_change} />
-                                    <label htmlFor={(spc.block ? spc.block.Device : spc.desc) + "-row-checkbox"}
-                                           className='data-list-row-checkbox-label'>
-                                        <DataListItemCells
+                        const datalistcells = (
+                            <DataListItemCells
                                             dataListCells={[
                                                 <DataListCell key="select-space-name" className="select-space-name">
                                                     {format_size_and_text(spc.size, desc)}
@@ -737,7 +788,18 @@ export const SelectSpaces = (tag, title, options) => {
                                                     {block}
                                                 </DataListCell>,
                                             ]}
-                                        />
+                            />);
+
+                        return (
+                            <DataListItem key={spc.block ? spc.block.Device : spc.desc}>
+                                <DataListItemRow>
+                                    <DataListCheck id={(spc.block ? spc.block.Device : spc.desc) + "-row-checkbox"}
+                                                   isDisabled={options.min_selected &&
+                                                               selected && val.length <= options.min_selected}
+                                                   isChecked={selected} onChange={on_change} />
+                                    <label htmlFor={(spc.block ? spc.block.Device : spc.desc) + "-row-checkbox"}
+                                           className='data-list-row-checkbox-label'>
+                                        {datalistcells}
                                     </label>
                                 </DataListItemRow>
                             </DataListItem>
@@ -923,7 +985,7 @@ class SizeSliderElement extends React.Component {
         const { unit } = this.state;
 
         const change_slider = (_event, f) => {
-            onChange(Math.max(min, size_slider_round(f * max / 100, round)));
+            onChange(Math.max(min, size_slider_round(f, round)));
         };
 
         const change_text = (value) => {
@@ -952,7 +1014,8 @@ class SizeSliderElement extends React.Component {
         return (
             <Grid hasGutter className="size-slider">
                 <GridItem span={12} sm={8}>
-                    <Slider showBoundaries={false} value={(slider_val / max) * 100} onChange={change_slider} />
+                    <Slider showBoundaries={false} min={min} max={max} step={(max - min) / 500}
+                            value={slider_val} onChange={change_slider} />
                 </GridItem>
                 <GridItem span={6} sm={2}>
                     <TextInputPF4 className="size-text" value={text_val} onChange={(_event, value) => change_text(value)} />
@@ -1045,18 +1108,17 @@ export const SizeSlider = (tag, title, options) => {
 export const BlockingMessage = (usage) => {
     const usage_desc = {
         pvol: _("physical volume of LVM2 volume group"),
-        mdraid: _("member of RAID device"),
+        "mdraid-member": _("member of MDRAID device"),
         vdo: _("backing device for VDO device"),
-        "stratis-pool-member": _("member of Stratis pool")
+        "stratis-pool-member": _("member of Stratis pool"),
+        mounted: _("Filesystem outside the target"),
+        "btrfs-device": _("device of btrfs volume"),
     };
 
     const rows = [];
     usage.forEach(use => {
         if (use.blocking && use.block) {
-            const fsys = client.blocks_stratis_fsys[use.block.path];
-            const name = (fsys
-                ? fsys.Devnode
-                : block_name(client.blocks[use.block.CryptoBackingDevice] || use.block));
+            const name = teardown_block_name(use);
             rows.push({
                 columns: [name, use.location || "-", usage_desc[use.usage] || "-"]
             });
@@ -1087,13 +1149,13 @@ const UsersPopover = ({ users }) => {
             bodyContent={
                 <>
                     { services.length > 0
-                        ? <p>
-                            <b>{_("Services using the location")}</b>
+                        ? <>
+                            <p><b>{_("Services using the location")}</b></p>
                             <List>
                                 { services.slice(0, max).map((u, i) => <ListItem key={i}>{u.unit.replace(/\.service$/, "")}</ListItem>) }
                                 { services.length > max ? <ListItem key={max}>...</ListItem> : null }
                             </List>
-                        </p>
+                        </>
                         : null
                     }
                     { services.length > 0 && processes.length > 0
@@ -1101,13 +1163,13 @@ const UsersPopover = ({ users }) => {
                         : null
                     }
                     { processes.length > 0
-                        ? <p>
-                            <b>{_("Processes using the location")}</b>
+                        ? <>
+                            <p><b>{_("Processes using the location")}</b></p>
                             <List>
                                 { processes.slice(0, max).map((u, i) => <ListItem key={i}>{u.comm} (user: {u.user}, pid: {u.pid})</ListItem>) }
                                 { processes.length > max ? <ListItem key={max}>...</ListItem> : null }
                             </List>
-                        </p>
+                        </>
                         : null
                     }
                 </>}>
@@ -1118,20 +1180,46 @@ const UsersPopover = ({ users }) => {
         </Popover>);
 };
 
-export const TeardownMessage = (usage) => {
+function is_expected_unmount(usage, expect_single_unmount) {
+    return (expect_single_unmount && usage.length == 1 &&
+            usage[0].usage == "mounted" && usage[0].location == expect_single_unmount);
+}
+
+const teardown_block_name = use => {
+    const block_stratis = client.blocks_stratis_fsys[use.block.path];
+    const block_btrfs = client.blocks_fsys_btrfs[use.block.path];
+    let name;
+    if (block_stratis) {
+        name = block_stratis.Devnode;
+    } else if (block_btrfs && use.name) {
+        name = use.name;
+    } else {
+        name = block_name(client.blocks[use.block.CryptoBackingDevice] || use.block);
+    }
+
+    return name;
+};
+
+export const TeardownMessage = (usage, expect_single_unmount) => {
     if (usage.length == 0)
         return null;
 
+    if (is_expected_unmount(usage, expect_single_unmount))
+        return <StopProcessesMessage mount_point={expect_single_unmount} users={usage[0].users} />;
+
     const rows = [];
     usage.forEach((use, index) => {
         if (use.block) {
-            const fsys = client.blocks_stratis_fsys[use.block.path];
-            const name = (fsys
-                ? fsys.Devnode
-                : block_name(client.blocks[use.block.CryptoBackingDevice] || use.block));
+            const name = teardown_block_name(use);
+            let location = use.location;
+            if (use.usage == "mounted") {
+                location = client.strip_mount_point_prefix(location);
+                if (location === false)
+                    location = _("(Not part of target)");
+            }
             rows.push({
                 columns: [name,
-                    use.location || "-",
+                    location || "-",
                     use.actions.length ? use.actions.join(", ") : "-",
                     {
                         title: <UsersPopover users={use.users || []} />,
@@ -1156,7 +1244,25 @@ export const TeardownMessage = (usage) => {
         </div>);
 };
 
-export function init_active_usage_processes(client, usage) {
+export function teardown_danger_message(usage, expect_single_unmount) {
+    if (is_expected_unmount(usage, expect_single_unmount))
+        return stop_processes_danger_message(usage[0].users);
+
+    const usage_with_users = usage.filter(u => u.users);
+    const n_processes = usage_with_users.reduce((sum, u) => sum + u.users.filter(u => u.pid).length, 0);
+    const n_services = usage_with_users.reduce((sum, u) => sum + u.users.filter(u => u.unit).length, 0);
+    if (n_processes > 0 && n_services > 0) {
+        return _("Related processes and services will be forcefully stopped.");
+    } else if (n_processes > 0) {
+        return _("Related processes will be forcefully stopped.");
+    } else if (n_services > 0) {
+        return _("Related services will be forcefully stopped.");
+    } else {
+        return null;
+    }
+}
+
+export function init_active_usage_processes(client, usage, expect_single_unmount) {
     return {
         title: _("Checking related processes"),
         func: dlg => {
@@ -1169,22 +1275,19 @@ export function init_active_usage_processes(client, usage) {
                 } else
                     return Promise.resolve();
             }).then(() => {
-                dlg.set_attribute("Teardown", TeardownMessage(usage));
-                const usage_with_users = usage.filter(u => u.users);
-                const n_processes = usage_with_users.reduce((sum, u) => sum + u.users.filter(u => u.pid).length, 0);
-                const n_services = usage_with_users.reduce((sum, u) => sum + u.users.filter(u => u.unit).length, 0);
-                if (n_processes > 0 && n_services > 0)
-                    dlg.add_danger(_("Related processes and services will be forcefully stopped."));
-                else if (n_processes > 0)
-                    dlg.add_danger(_("Related processes will be forcefully stopped."));
-                else if (n_services > 0)
-                    dlg.add_danger(_("Related services will be forcefully stopped."));
+                dlg.set_attribute("Teardown", TeardownMessage(usage, expect_single_unmount));
+                const msg = teardown_danger_message(usage, expect_single_unmount);
+                if (msg)
+                    dlg.add_danger(msg);
             });
         }
     };
 }
 
 export const StopProcessesMessage = ({ mount_point, users }) => {
+    if (!users || users.length == 0)
+        return null;
+
     const process_rows = users.filter(u => u.pid).map(u => {
         return {
             columns: [
@@ -1214,7 +1317,8 @@ export const StopProcessesMessage = ({ mount_point, users }) => {
     return (
         <div className="modal-footer-teardown">
             { process_rows.length > 0
-                ? <p>{fmt_to_fragments(_("The mount point $0 is in use by these processes:"), <b>{mount_point}</b>)}
+                ? <>
+                    <p>{fmt_to_fragments(_("The mount point $0 is in use by these processes:"), <b>{mount_point}</b>)}</p>
                     <ListingTable variant='compact'
                                   columns={
                                       [
@@ -1225,7 +1329,7 @@ export const StopProcessesMessage = ({ mount_point, users }) => {
                                       ]
                                   }
                                       rows={process_rows} />
-                </p>
+                </>
                 : null
             }
             { process_rows.length > 0 && service_rows.length > 0
@@ -1233,7 +1337,8 @@ export const StopProcessesMessage = ({ mount_point, users }) => {
                 : null
             }
             { service_rows.length > 0
-                ? <p>{fmt_to_fragments(_("The mount point $0 is in use by these services:"), <b>{mount_point}</b>)}
+                ? <>
+                    <p>{fmt_to_fragments(_("The mount point $0 is in use by these services:"), <b>{mount_point}</b>)}</p>
                     <ListingTable variant='compact'
                                   columns={
                                       [
@@ -1244,7 +1349,7 @@ export const StopProcessesMessage = ({ mount_point, users }) => {
                                       ]
                                   }
                                   rows={service_rows} />
-                </p>
+                </>
                 : null
             }
         </div>);
