@@ -17,50 +17,41 @@
  * along with Cockpit; If not, see <http://www.gnu.org/licenses/>.
  */
 
-import cockpit from 'cockpit';
 import React, { useState, useEffect, useRef } from 'react';
-import { superuser } from "superuser";
-import { apply_modal_dialog } from "cockpit-components-dialog.jsx";
 
 import { Button } from "@patternfly/react-core/dist/esm/components/Button/index.js";
 import { Checkbox } from "@patternfly/react-core/dist/esm/components/Checkbox/index.js";
-import { Card, CardActions, CardBody, CardHeader, CardTitle } from "@patternfly/react-core/dist/esm/components/Card/index.js";
-import { EmptyState, EmptyStateIcon, EmptyStateSecondaryActions, EmptyStateVariant } from "@patternfly/react-core/dist/esm/components/EmptyState/index.js";
-import { Flex } from "@patternfly/react-core/dist/esm/layouts/Flex/index.js";
+import { Card, CardBody, CardHeader, CardTitle } from '@patternfly/react-core/dist/esm/components/Card/index.js';
+import { EmptyState, EmptyStateActions, EmptyStateFooter, EmptyStateHeader, EmptyStateIcon, EmptyStateVariant } from "@patternfly/react-core/dist/esm/components/EmptyState/index.js";
+import { Flex, FlexItem } from "@patternfly/react-core/dist/esm/layouts/Flex/index.js";
 import { HelperText, HelperTextItem } from "@patternfly/react-core/dist/esm/components/HelperText/index.js";
-import { Label } from "@patternfly/react-core/dist/esm/components/Label/index.js";
-import { LabelGroup } from "@patternfly/react-core/dist/esm/components/LabelGroup/index.js";
-import { Page, PageSection } from "@patternfly/react-core/dist/esm/components/Page/index.js";
+import { Label, LabelGroup } from "@patternfly/react-core/dist/esm/components/Label/index.js";
+import { Page, PageBreadcrumb, PageSection } from "@patternfly/react-core/dist/esm/components/Page/index.js";
 import { Gallery } from "@patternfly/react-core/dist/esm/layouts/Gallery/index.js";
-import { Select, SelectOption, SelectVariant } from "@patternfly/react-core/dist/esm/components/Select/index.js";
-import { Text, TextVariants } from "@patternfly/react-core/dist/esm/components/Text/index.js";
+import { Select, SelectOption } from "@patternfly/react-core/dist/esm/deprecated/components/Select/index.js";
 import { Breadcrumb, BreadcrumbItem } from "@patternfly/react-core/dist/esm/components/Breadcrumb/index.js";
 import { Form, FormGroup } from "@patternfly/react-core/dist/esm/components/Form/index.js";
 import { TextInput } from "@patternfly/react-core/dist/esm/components/TextInput/index.js";
 import { Spinner } from "@patternfly/react-core/dist/esm/components/Spinner/index.js";
-import { Title } from "@patternfly/react-core/dist/esm/components/Title/index.js";
 import { Popover } from "@patternfly/react-core/dist/esm/components/Popover/index.js";
 import { ExclamationCircleIcon, HelpIcon, UndoIcon } from '@patternfly/react-icons';
+
+import cockpit from 'cockpit';
+import { superuser } from "superuser";
+import * as timeformat from "timeformat.js";
+import { apply_modal_dialog } from "cockpit-components-dialog.jsx";
+
 import { show_unexpected_error } from "./dialog-utils.js";
 import { delete_account_dialog } from "./delete-account-dialog.js";
 import { account_expiration_dialog, password_expiration_dialog } from "./expiration-dialogs.js";
+import { account_shell_dialog } from "./shell-dialog.js";
 import { set_password_dialog, reset_password_dialog } from "./password-dialogs.js";
 import { AccountLogs } from "./account-logs-panel.jsx";
 import { AuthorizedKeys } from "./authorized-keys-panel.js";
-import * as timeformat from "timeformat.js";
+import { get_locked } from "./utils.js";
 
 const _ = cockpit.gettext;
 
-function get_locked(name) {
-    return cockpit.spawn(["passwd", "-S", name], { environ: ["LC_ALL=C"], superuser: "require" })
-            .catch(() => "")
-            .then(content => {
-                const status = content.split(" ")[1];
-                // libuser uses "LK", shadow-utils use "L".
-                return status && (status == "LK" || status == "L");
-            });
-}
-
 function get_expire(name) {
     function parse_expire(data) {
         let account_expiration = '';
@@ -105,7 +96,7 @@ function get_expire(name) {
             .then(parse_expire);
 }
 
-export function AccountDetails({ accounts, groups, shadow, current_user, user }) {
+export function AccountDetails({ accounts, groups, current_user, user, shells }) {
     const [expiration, setExpiration] = useState(null);
     useEffect(() => {
         get_expire(user).then(setExpiration);
@@ -116,7 +107,7 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
             get_expire(user).then(setExpiration);
         });
         return handle.close;
-    }, [user, accounts, shadow]);
+    }, [user, accounts]);
 
     const [edited_real_name, set_edited_real_name] = useState(null);
     const [committing_real_name, set_committing_real_name] = useState(false);
@@ -155,7 +146,6 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
                                this is a workaround for different ways of handling a locked account
                                https://github.com/cockpit-project/cockpit/issues/1216
                                https://bugzilla.redhat.com/show_bug.cgi?id=853153
-                               This seems to be fixed in fedora 23 (usermod catches the different locking behavior)
                             */
                                 if (locked != value && !dont_retry_if_stuck) {
                                     console.log("Account locked state doesn't match desired value, trying again.");
@@ -182,11 +172,9 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
 
     if (!accounts.length) {
         return (
-            <EmptyState variant={EmptyStateVariant.small}>
-                <Spinner isSVG size="xl" />
-                <Title headingLevel="h1" size="lg">
-                    {_("Loading...")}
-                </Title>
+            <EmptyState variant={EmptyStateVariant.sm}>
+                <EmptyStateHeader titleText={_("Loading...")} headingLevel="h1" />
+                <EmptyStateFooter><Spinner size="xl" /></EmptyStateFooter>
             </EmptyState>
         );
     }
@@ -195,16 +183,15 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
 
     if (!account) {
         return (
-            <EmptyState variant={EmptyStateVariant.small} id="account-failure">
-                <EmptyStateIcon icon={ExclamationCircleIcon} />
-                <Title headingLevel="h1" size="lg">
-                    {_("Account not available or cannot be edited.")}
-                </Title>
-                <EmptyStateSecondaryActions>
-                    <Breadcrumb>
-                        <BreadcrumbItem to="#/">{_("Back to accounts")}</BreadcrumbItem>
-                    </Breadcrumb>
-                </EmptyStateSecondaryActions>
+            <EmptyState variant={EmptyStateVariant.sm} id="account-failure">
+                <EmptyStateHeader titleText={<>{_("Account not available or cannot be edited.")}</>} icon={<EmptyStateIcon icon={ExclamationCircleIcon} />} headingLevel="h1" />
+                <EmptyStateFooter>
+                    <EmptyStateActions>
+                        <Breadcrumb>
+                            <BreadcrumbItem to="#/">{_("Back to accounts")}</BreadcrumbItem>
+                        </Breadcrumb>
+                    </EmptyStateActions>
+                </EmptyStateFooter>
             </EmptyState>
         );
     }
@@ -228,33 +215,33 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
     else
         last_login = timeformat.dateTime(new Date(account.lastLogin));
 
+    const actions = superuser.allowed && (
+        <>
+            <Button variant="secondary" onClick={() => logout_account()} id="account-logout"
+              isDisabled={!account.loggedIn || account.uid == 0}>
+                {_("Terminate session")}
+            </Button>
+            { "\n" }
+            <Button isDisabled={account.uid == 0} variant="danger" id="account-delete"
+                  onClick={() => delete_account_dialog(account)}>
+                {_("Delete")}
+            </Button>
+        </>
+    );
+
     return (
-        <Page groupProps={{ sticky: 'top' }}
-              isBreadcrumbGrouped
-              id="account"
-              breadcrumb={
-                  <Breadcrumb>
-                      <BreadcrumbItem to="#/">{_("Accounts")}</BreadcrumbItem>
-                      <BreadcrumbItem isActive>{title_name}</BreadcrumbItem>
-                  </Breadcrumb>}>
+        <Page id="account">
+            <PageBreadcrumb stickyOnBreakpoint={{ default: "top" }}>
+                <Breadcrumb>
+                    <BreadcrumbItem to="#/">{_("Accounts")}</BreadcrumbItem>
+                    <BreadcrumbItem isActive>{title_name}</BreadcrumbItem>
+                </Breadcrumb>
+            </PageBreadcrumb>
             <PageSection>
                 <Gallery hasGutter>
                     <Card className="account-details" id="account-details">
-                        <CardHeader>
-                            <CardTitle id="account-title"><Text component={TextVariants.h2}>{title_name}</Text></CardTitle>
-                            { superuser.allowed &&
-                            <CardActions>
-                                <Button variant="secondary" onClick={() => logout_account()} id="account-logout"
-                                  isDisabled={!account.loggedIn || account.uid == 0}>
-                                    {_("Terminate session")}
-                                </Button>
-                                { "\n" }
-                                <Button isDisabled={account.uid == 0} variant="danger" id="account-delete"
-                                      onClick={() => delete_account_dialog(account)}>
-                                    {_("Delete")}
-                                </Button>
-                            </CardActions>
-                            }
+                        <CardHeader actions={{ actions }}>
+                            <CardTitle id="account-title" component="h2">{title_name}</CardTitle>
                         </CardHeader>
                         <CardBody>
                             <Form isHorizontal onSubmit={apply_modal_dialog}>
@@ -263,12 +250,12 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
                                         ? <TextInput id="account-real-name"
                                                      isDisabled={committing_real_name || account.uid == 0}
                                                      value={edited_real_name !== null ? edited_real_name : account.gecos}
-                                                     onKeyPress={event => {
+                                                     onKeyDown={event => {
                                                          if (event.key == "Enter") {
                                                              event.target.blur();
                                                          }
                                                      }}
-                                                     onChange={value => set_edited_real_name(value)}
+                                                     onChange={(_event, value) => set_edited_real_name(value)}
                                                      onBlur={() => change_real_name()} />
                                         : <output id="account-real-name">{account.gecos}</output>}
                                 </FormGroup>
@@ -280,68 +267,77 @@ export function AccountDetails({ accounts, groups, shadow, current_user, user })
                                     <output id="account-last-login">{last_login}</output>
                                 </FormGroup>
                                 <FormGroup fieldId="account-locked" label={_("Options")} hasNoPaddingTop>
-                                    <div>
-                                        <div className="account-column-one">
-                                            <Flex spaceItems={{ default: 'spaceItemsSm' }} alignItems={{ default: 'alignItemsCenter' }}>
-                                                <Checkbox id="account-locked"
-                                                          isDisabled={!superuser.allowed || edited_locked != null || user == current_user}
-                                                          isChecked={edited_locked != null ? edited_locked : account.isLocked}
-                                                          onChange={checked => change_locked(checked)}
-                                                          label={_("Disallow interactive password")} />
+                                    <Flex spaceItems={{ default: 'spaceItemsSm' }} alignItems={{ default: 'alignItemsCenter' }}>
+                                        <FlexItem spacer={{ default: 'spacerNone' }}>
+                                            <Checkbox id="account-locked"
+                                                        isDisabled={!superuser.allowed || edited_locked != null || user == current_user}
+                                                        isChecked={edited_locked != null ? edited_locked : account.isLocked}
+                                                        onChange={(_event, checked) => change_locked(checked)}
+                                                        label={_("Disallow interactive password")} />
+                                        </FlexItem>
 
-                                                <Popover bodyContent={_("Other authentication methods are still available even when interactive password authentication is not allowed.")}
-                                                         showClose={false}>
-                                                    <HelpIcon />
-                                                </Popover>
-                                            </Flex>
-                                        </div>
-                                        <Flex flex={{ default: 'inlineFlex' }}>
-                                            <span id="account-expiration-text">
-                                                {expiration.account_text}
-                                            </span>
-                                            <Button onClick={() => account_expiration_dialog(account, expiration.account_date)}
-                                                    isDisabled={!superuser.allowed}
-                                                    variant="link"
-                                                    isInline
-                                                    id="account-expiration-button">
-                                                {_("edit")}
-                                            </Button>
-                                        </Flex>
-                                    </div>
+                                        <Popover bodyContent={_("Other authentication methods are still available even when interactive password authentication is not allowed.")}
+                                                    showClose={false}>
+                                            <HelpIcon />
+                                        </Popover>
+                                        <span id="account-expiration-text">
+                                            {expiration.account_text}
+                                        </span>
+                                        <Button onClick={() => account_expiration_dialog(account, expiration.account_date)}
+                                                isDisabled={!superuser.allowed}
+                                                variant="link"
+                                                isInline
+                                                id="account-expiration-button">
+                                            {_("edit")}
+                                        </Button>
+                                    </Flex>
                                 </FormGroup>
                                 { self_mod_allowed &&
                                 <FormGroup fieldId="account-set-password" label={_("Password")}>
-                                    <div>
-                                        <div className="account-column-one">
-                                            { self_mod_allowed &&
-                                            <Button variant="secondary" id="account-set-password"
-                                      onClick={() => set_password_dialog(account, current_user)}>
-                                                {_("Set password")}
-                                            </Button>
-                                            }
-                                            { "\n" }
-                                            { superuser.allowed &&
-                                            <Button variant="secondary" id="password-reset-button"
-                                              onClick={() => reset_password_dialog(account)}>
-                                                {_("Force change")}
-                                            </Button>
-                                            }
-                                        </div>
-                                        <Flex flex={{ default: 'inlineFlex' }}>
-                                            <span id="password-expiration-text">
-                                                {expiration.password_text}
-                                            </span>
-                                            <Button onClick={() => password_expiration_dialog(account, expiration.password_days)}
-                                                    isDisabled={!superuser.allowed}
-                                                    variant="link"
-                                                    isInline
-                                                    id="password-expiration-button">
-                                                {_("edit")}
-                                            </Button>
-                                        </Flex>
+                                    <div className="account-column-one">
+                                        { self_mod_allowed &&
+                                        <Button variant="secondary" id="account-set-password"
+                                    onClick={() => set_password_dialog(account, current_user)}>
+                                            {_("Set password")}
+                                        </Button>
+                                        }
+                                        { "\n" }
+                                        { superuser.allowed &&
+                                        <Button variant="secondary" id="password-reset-button"
+                                            onClick={() => reset_password_dialog(account)}>
+                                            {_("Force change")}
+                                        </Button>
+                                        }
                                     </div>
+                                    <Flex flex={{ default: 'inlineFlex' }}>
+                                        <span id="password-expiration-text">
+                                            {expiration.password_text}
+                                        </span>
+                                        <Button onClick={() => password_expiration_dialog(account, expiration.password_days)}
+                                                isDisabled={!superuser.allowed}
+                                                variant="link"
+                                                isInline
+                                                id="password-expiration-button">
+                                            {_("edit")}
+                                        </Button>
+                                    </Flex>
                                 </FormGroup>
                                 }
+                                { account.home && <FormGroup fieldId="account-home-dir" hasNoPaddingTop label={_("Home directory")}>
+                                    <output id="account-home-dir">{account.home}</output>
+                                </FormGroup> }
+                                { account.shell && <FormGroup fieldId="account-shell" hasNoPaddingTop label={_("Shell")}>
+                                    <Flex flex={{ default: 'inlineFlex' }}>
+                                        <output id="account-shell">{account.shell}</output>
+                                        <Button onClick={() => account_shell_dialog(account, shells)}
+                                                isDisabled={!superuser.allowed}
+                                                variant="link"
+                                                isInline
+                                                id="change-shell-button">
+                                            {_("change")}
+                                        </Button>
+                                    </Flex>
+                                </FormGroup> }
                             </Form>
                         </CardBody>
                     </Card>
@@ -364,7 +360,7 @@ export const AccountGroupsSelect = ({ name, loggedIn, groups, setError }) => {
     useEffect(() => {
         const usedGroups = groups.filter(group => group.userlist.includes(name));
         const primaryGroup = groups.find(group => group.userlistPrimary.includes(name));
-        const _primaryGroupName = primaryGroup ? primaryGroup.name : undefined;
+        const _primaryGroupName = primaryGroup?.name;
         const _selected = usedGroups.map(group => group.name);
         if (primaryGroup)
             _selected.push(_primaryGroupName);
@@ -443,16 +439,6 @@ export const AccountGroupsSelect = ({ name, loggedIn, groups, setError }) => {
     return (
         <FormGroup
             fieldId="account-groups"
-            helperText={
-                (history.length > 0)
-                    ? <HelperText className="pf-c-form__helper-text">
-                        <Flex>
-                            {loggedIn && <HelperTextItem id="account-groups-helper" variant="warning">{_("The user must log out and log back in for the new configuration to take effect.")}</HelperTextItem>}
-                            {history.length > 0 && <Button variant="link" id="group-undo-btn" isInline icon={<UndoIcon />} onClick={undoGroupChanges}>{_("Undo")}</Button>}
-                        </Flex>
-                    </HelperText>
-                    : ''
-            }
             id="account-groups-form-group"
             label={_("Groups")}
             validated={history.length > 0 ? "warning" : "default"}
@@ -463,10 +449,10 @@ export const AccountGroupsSelect = ({ name, loggedIn, groups, setError }) => {
                    isDisabled={!superuser.allowed || loading}
                    isOpen={isOpenGroup}
                    onSelect={onSelectGroup}
-                   onToggle={setIsOpenGroup}
+                   onToggle={(_, isOpen) => setIsOpenGroup(isOpen)}
                    selections={selected}
                    toggleId="account-groups"
-                   variant={SelectVariant.typeaheadMulti}
+                   variant="typeaheadmulti"
                 >
                     {groups.map((option, index) => (
                         <SelectOption
@@ -477,6 +463,15 @@ export const AccountGroupsSelect = ({ name, loggedIn, groups, setError }) => {
                     ))}
                 </Select>
                 : chipGroupComponent()}
+            {(history.length > 0)
+                ? <HelperText className="pf-v5-c-form__helper-text">
+                    <Flex>
+                        {loggedIn && <HelperTextItem id="account-groups-helper" variant="warning">{_("The user must log out and log back in for the new configuration to take effect.")}</HelperTextItem>}
+                        {history.length > 0 && <Button variant="link" id="group-undo-btn" isInline icon={<UndoIcon />} onClick={undoGroupChanges}>{_("Undo")}</Button>}
+                    </Flex>
+                </HelperText>
+                : ''
+            }
         </FormGroup>
     );
 };
