@@ -17,14 +17,14 @@
  * along with Cockpit; If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include "config.h"
+
 /* This gets logged as part of the (more verbose) protocol logging */
 #ifdef G_LOG_DOMAIN
 #undef G_LOG_DOMAIN
 #endif
 #define G_LOG_DOMAIN "cockpit-protocol"
 
-#include "config.h"
-
 #include "cockpitwebresponse.h"
 #include "cockpitwebfilter.h"
 
@@ -38,14 +38,6 @@
 #include <stdlib.h>
 #include <string.h>
 
-/**
- * Certain processes may want to have a non-default error page.
- */
-const gchar *cockpit_web_failure_resource = NULL;
-
-static const gchar default_failure_template[] =
-  "<html><head><title>@@message@@</title></head><body>@@message@@</body></html>\n";
-
 /**
  * CockpitWebResponse:
  *
@@ -64,12 +56,11 @@ struct _CockpitWebResponse {
   const gchar *logname;
   const gchar *path;
   gchar *full_path;
-  gchar *query;
   gchar *url_root;
   gchar *method;
   gchar *origin;
 
-  CockpitWebResponseFlags flags;
+  gchar *protocol;
   CockpitCacheType cache_type;
 
   /* The output queue */
@@ -164,8 +155,8 @@ cockpit_web_response_finalize (GObject *object)
 {
   CockpitWebResponse *self = COCKPIT_WEB_RESPONSE (object);
 
+  g_free (self->protocol);
   g_free (self->full_path);
-  g_free (self->query);
   g_free (self->url_root);
   g_free (self->method);
   g_free (self->origin);
@@ -195,7 +186,6 @@ cockpit_web_response_class_init (CockpitWebResponseClass *klass)
  * cockpit_web_response_new:
  * @io: the stream to send on
  * @path: the path resource or NULL
- * @query: the query string or NULL
  * @in_headers: input headers or NULL
  * @flags: in #COCKPIT_WEB_RESPONSE_FOR_TLS_PROXY mode, the origin is assumed to
  *         be https://<host> even for a non-HTTPS connection.
@@ -212,14 +202,13 @@ CockpitWebResponse *
 cockpit_web_response_new (GIOStream *io,
                           const gchar *original_path,
                           const gchar *path,
-                          const gchar *query,
                           GHashTable *in_headers,
-                          CockpitWebResponseFlags flags)
+                          const gchar *method,
+                          const gchar *protocol)
 {
   CockpitWebResponse *self;
   GOutputStream *out;
   const gchar *connection;
-  const gchar *protocol = NULL;
   const gchar *host = NULL;
   gint offset;
 
@@ -245,6 +234,7 @@ cockpit_web_response_new (GIOStream *io,
   self->url_root = NULL;
   self->full_path = g_strdup (path);
   self->path = self->full_path;
+  cockpit_web_response_set_method (self, method);
 
   if (path && original_path)
     {
@@ -253,7 +243,6 @@ cockpit_web_response_new (GIOStream *io,
         self->url_root = g_strndup (original_path, offset);
     }
 
-  self->query = g_strdup (query);
   if (self->path)
     self->logname = self->path;
   else
@@ -268,10 +257,9 @@ cockpit_web_response_new (GIOStream *io,
       host = g_hash_table_lookup (in_headers, "Host");
     }
 
-  self->flags = flags;
-  protocol = cockpit_web_response_get_protocol (self, in_headers);
-  if (protocol && host)
-    self->origin = g_strdup_printf ("%s://%s", protocol, host);
+  self->protocol = g_strdup (protocol ?: "http");
+  if (host)
+    self->origin = g_strdup_printf ("%s://%s", self->protocol, host);
 
   return self;
 }
@@ -281,6 +269,7 @@ cockpit_web_response_set_method (CockpitWebResponse *response,
                                  const gchar *method)
 {
   g_return_if_fail (g_strcmp0 (method, "GET") == 0 || g_strcmp0 (method, "HEAD") == 0);
+  g_free (response->method);
   response->method = g_strdup (method);
 }
 
@@ -308,19 +297,6 @@ cockpit_web_response_get_url_root (CockpitWebResponse *self) {
   return self->url_root;
 }
 
-/**
- * cockpit_web_response_get_query:
- * @self: the response
- *
- * Returns: the resource path for response
- */
-const gchar *
-cockpit_web_response_get_query (CockpitWebResponse *self)
-{
-  g_return_val_if_fail (COCKPIT_IS_WEB_RESPONSE (self), NULL);
-  return self->query;
-}
-
 /**
  * cockpit_web_response_get_stream:
  * @self: the response
@@ -747,6 +723,7 @@ enum {
     HEADER_REFERRER_POLICY = 1 << 5,
     HEADER_CONTENT_TYPE_OPTIONS = 1 << 6,
     HEADER_CROSS_ORIGIN_RESOURCE_POLICY = 1 << 7,
+    HEADER_X_FRAME_OPTIONS = 1 << 8,
 };
 
 static GString *
@@ -789,6 +766,8 @@ append_header (GString *string,
     return HEADER_CONTENT_TYPE_OPTIONS;
   if (g_ascii_strcasecmp ("Cross-Origin-Resource-Policy", name) == 0)
     return HEADER_CROSS_ORIGIN_RESOURCE_POLICY;
+  if (g_ascii_strcasecmp ("X-Frame-Options", name) == 0)
+    return HEADER_X_FRAME_OPTIONS;
   if (g_ascii_strcasecmp ("Content-Length", name) == 0 ||
       g_ascii_strcasecmp ("Transfer-Encoding", name) == 0 ||
       g_ascii_strcasecmp ("Connection", name) == 0)
@@ -872,16 +851,14 @@ finish_headers (CockpitWebResponse *self,
 
   if ((seen & HEADER_CACHE_CONTROL) == 0 && status >= 200 && status <= 299)
     {
-      if (self->cache_type == COCKPIT_WEB_RESPONSE_CACHE_FOREVER)
-        g_string_append (string, "Cache-Control: max-age=31556926, public\r\n");
-      else if (self->cache_type == COCKPIT_WEB_RESPONSE_NO_CACHE)
+      if (self->cache_type == COCKPIT_WEB_RESPONSE_NO_CACHE)
         g_string_append (string, "Cache-Control: no-cache, no-store\r\n");
-      else if (self->cache_type == COCKPIT_WEB_RESPONSE_CACHE_PRIVATE)
+      else if (self->cache_type == COCKPIT_WEB_RESPONSE_CACHE)
         g_string_append (string, "Cache-Control: max-age=86400, private\r\n");
     }
 
   if ((seen & HEADER_VARY) == 0 && status >= 200 && status <= 299 &&
-      self->cache_type == COCKPIT_WEB_RESPONSE_CACHE_PRIVATE)
+      self->cache_type == COCKPIT_WEB_RESPONSE_CACHE)
     {
       g_string_append (string, "Vary: Cookie\r\n");
     }
@@ -900,6 +877,9 @@ finish_headers (CockpitWebResponse *self,
    * be able to read any resource. This does *not* affect embedding with <iframe> */
   if ((seen & HEADER_CROSS_ORIGIN_RESOURCE_POLICY) == 0)
     g_string_append (string, "Cross-Origin-Resource-Policy: same-origin\r\n");
+  /* This is the counterpart for iframe embedding, line of defence against clickjacking */
+  if ((seen & HEADER_X_FRAME_OPTIONS) == 0)
+    g_string_append (string, "X-Frame-Options: sameorigin\r\n");
 
   g_string_append (string, "\r\n");
   return g_string_free_to_bytes (string);
@@ -908,7 +888,7 @@ finish_headers (CockpitWebResponse *self,
 /**
  * cockpit_web_response_set_cache_type:
  * @self: the response
- * @cache_type: Ensures the apropriate cache headers are returned for
+ * @cache_type: Ensures the appropriate cache headers are returned for
    the given cache type.
  */
 void
@@ -1112,12 +1092,9 @@ cockpit_web_response_error (CockpitWebResponse *self,
                             ...)
 {
   va_list var_args;
-  gchar *reason = NULL;
-  gchar *escaped = NULL;
+  g_autofree gchar *reason = NULL;
+  g_autofree gchar *escaped = NULL;
   const gchar *message;
-  GBytes *input = NULL;
-  GList *output, *l;
-  GError *error = NULL;
 
   g_return_if_fail (COCKPIT_IS_WEB_RESPONSE (self));
 
@@ -1172,22 +1149,6 @@ cockpit_web_response_error (CockpitWebResponse *self,
 
   g_debug ("%s: returning error: %u %s", self->logname, code, message);
 
-  if (cockpit_web_failure_resource)
-    {
-      input = g_resources_lookup_data (cockpit_web_failure_resource, G_RESOURCE_LOOKUP_FLAGS_NONE, &error);
-      if (input == NULL)
-        {
-          g_critical ("couldn't load: %s: %s", cockpit_web_failure_resource, error->message);
-          g_error_free (error);
-        }
-    }
-
-  if (!input)
-    input = g_bytes_new_static (default_failure_template, strlen (default_failure_template));
-  output = cockpit_template_expand (input, substitute_message,
-                                    "@@", "@@", (gpointer)message);
-  g_bytes_unref (input);
-
   /* If sending arbitrary messages, make sure they're escaped */
   if (reason)
     {
@@ -1207,17 +1168,21 @@ cockpit_web_response_error (CockpitWebResponse *self,
       cockpit_web_response_headers (self, code, message, -1, "Content-Type", "text/html; charset=utf8", NULL);
     }
 
-  for (l = output; l != NULL; l = g_list_next (l))
+  if (g_str_equal (self->method, "GET"))
     {
-      if (!cockpit_web_response_queue (self, l->data))
-        break;
-    }
-  if (l == NULL)
-    cockpit_web_response_complete (self);
-  g_list_free_full (output, (GDestroyNotify)g_bytes_unref);
+      extern const char *cockpit_webresponse_fail_html_text;
+      g_autoptr(GBytes) input = g_bytes_new_static (cockpit_webresponse_fail_html_text, strlen (cockpit_webresponse_fail_html_text));
+      g_autolist(GBytes) output = cockpit_template_expand (input, "@@", "@@", substitute_message, (gpointer) message);
 
-  g_free (reason);
-  g_free (escaped);
+      for (GList *l = output; l != NULL; l = g_list_next (l))
+        {
+          if (!cockpit_web_response_queue (self, l->data))
+            /* error: early exit */
+            return;
+        }
+    }
+
+  cockpit_web_response_complete (self);
 }
 
 /**
@@ -1298,24 +1263,11 @@ static void
 web_response_file (CockpitWebResponse *response,
                    const gchar *escaped,
                    const gchar **roots,
+                   gboolean search_gzip,
+                   gboolean accept_gzip,
                    CockpitTemplateFunc template_func,
                    gpointer user_data)
 {
-  const gchar *default_policy = "default-src 'self' 'unsafe-inline';";
-
-  const gchar *headers[5] = { NULL };
-  GError *error = NULL;
-  gchar *unescaped = NULL;
-  gchar *path = NULL;
-  gchar *alloc = NULL;
-  GMappedFile *file = NULL;
-  const gchar *root;
-  GBytes *body;
-  GList *output = NULL;
-  GList *l = NULL;
-  gint content_length = -1;
-  gint at = 0;
-
   g_return_if_fail (COCKPIT_IS_WEB_RESPONSE (response));
 
   if (!escaped)
@@ -1324,76 +1276,109 @@ web_response_file (CockpitWebResponse *response,
   g_return_if_fail (escaped != NULL);
 
   /* Someone is trying to escape the root directory, or access hidden files? */
-  unescaped = g_uri_unescape_string (escaped, "/");
+  g_autofree gchar *unescaped = g_uri_unescape_string (escaped, "/");
   if (!unescaped || strstr (unescaped, "/.") || strstr (unescaped, "../") || strstr (unescaped, "//"))
     {
       g_debug ("%s: invalid path request", escaped);
       cockpit_web_response_error (response, 404, NULL, "Not Found");
-      goto out;
+      return;
     }
 
-again:
-  root = *(roots++);
-  if (root == NULL)
+  gboolean is_gzip = FALSE;
+  g_autoptr(GMappedFile) file = NULL;
+  for (gint i = 0; roots[i]; i++)
     {
-      cockpit_web_response_error (response, 404, NULL, "Not Found");
-      goto out;
-    }
+      const gchar *root = roots[i];
+      g_autofree gchar *path = g_build_filename (root, unescaped, NULL);
 
-  g_free (path);
-  path = g_build_filename (root, unescaped, NULL);
+      if (g_file_test (path, G_FILE_TEST_IS_DIR))
+        {
+          cockpit_web_response_error (response, 403, NULL, "Directory Listing Denied");
+          return;
+        }
 
-  if (g_file_test (path, G_FILE_TEST_IS_DIR))
-    {
-      cockpit_web_response_error (response, 403, NULL, "Directory Listing Denied");
-      goto out;
-    }
+      /* As a double check of above behavior */
+      g_assert (path_has_prefix (path, root));
 
-  /* As a double check of above behavior */
-  g_assert (path_has_prefix (path, root));
+      g_autoptr(GError) error = NULL;
+      file = g_mapped_file_new (path, FALSE, &error);
+
+      if (file == NULL && search_gzip &&
+          g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
+        {
+          g_debug ("%s: file not found in root: %s, looking for .gz", escaped, root);
+          g_clear_error (&error);
+          g_autofree gchar *old_path = g_steal_pointer (&path);
+          path = g_strconcat (old_path, ".gz", NULL);
+          file = g_mapped_file_new (path, FALSE, &error);
+          is_gzip = file != NULL;
+        }
+
+      if (file != NULL)
+        break;
 
-  g_clear_error (&error);
-  file = g_mapped_file_new (path, FALSE, &error);
-  if (file == NULL)
-    {
       if (g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT) ||
           g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NAMETOOLONG))
         {
           g_debug ("%s: file not found in root: %s", escaped, root);
-          goto again;
         }
       else if (g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_PERM) ||
                g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_ACCES) ||
                g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_ISDIR))
         {
           cockpit_web_response_error (response, 403, NULL, "Access denied");
-          goto out;
+          return;
         }
       else
         {
           g_warning ("%s: %s", path, error->message);
           cockpit_web_response_error (response, 500, NULL, "Internal server error");
-          goto out;
+          return;
         }
     }
 
-  body = g_mapped_file_get_bytes (file);
+  if (file == NULL)
+    {
+      cockpit_web_response_error (response, 404, NULL, "Not Found");
+      return;
+    }
+
+  g_autoptr(GBytes) body = g_mapped_file_get_bytes (file);
+
+  if (is_gzip && (!accept_gzip || template_func))
+    {
+      /* We have gzipped content, but the client won't accept it, or
+       * template expansion was requested.  Decompress.
+       */
+      g_autoptr(GError) error = NULL;
+      g_autoptr(GBytes) body_gz = g_steal_pointer (&body);
+      body = cockpit_web_response_gunzip (body_gz, &error);
+      if (body == NULL)
+        {
+          g_warning ("%s", error->message);
+          cockpit_web_response_error (response, 500, NULL, "Internal server error");
+          return;
+        }
+      is_gzip = FALSE;
+    }
+
+  GList *output;
+  gint content_length = -1;
   if (template_func)
     {
-      output = cockpit_template_expand (body, template_func, "${", "}", user_data);
+      output = cockpit_template_expand (body, "${", "}", template_func, user_data);
     }
   else
     {
-      output = g_list_prepend (output, g_bytes_ref (body));
+      output = g_list_prepend (NULL, g_bytes_ref (body));
       content_length = g_bytes_get_size (body);
     }
-  g_bytes_unref (body);
+
+  GString *string = begin_headers (response, 200, "OK");
+  guint seen = 0;
 
   if (response->origin)
-    {
-      headers[at++] = "Access-Control-Allow-Origin";
-      headers[at++] = response->origin;
-    }
+    seen |= append_header (string, "Access-Control-Allow-Origin", response->origin);
 
   /*
    * The default Content-Security-Policy for .html files allows
@@ -1402,13 +1387,18 @@ again:
    */
   if (g_str_has_suffix (unescaped, ".html"))
     {
-      headers[at++] = "Content-Security-Policy";
-      headers[at++] = alloc = cockpit_web_response_security_policy (default_policy, response->origin);
+      const gchar *default_policy = "default-src 'self' 'unsafe-inline';";
+      g_autofree gchar *policy = cockpit_web_response_security_policy (default_policy, response->origin);
+      seen |= append_header (string, "Content-Security-Policy", policy);
     }
 
-  cockpit_web_response_headers (response, 200, "OK", content_length,
-                                headers[0], headers[1], headers[2], headers[3], NULL);
+  if (is_gzip)
+    seen |= append_header (string, "Content-Encoding", "gzip");
 
+  g_autoptr(GBytes) headers_block = finish_headers (response, string, content_length, 200, seen);
+  queue_bytes (response, headers_block);
+
+  GList *l;
   for (l = output; l != NULL; l = g_list_next (l))
     {
       if (!cockpit_web_response_queue (response, l->data))
@@ -1417,16 +1407,7 @@ again:
   if (l == NULL)
     cockpit_web_response_complete (response);
 
-out:
-  g_free (alloc);
-  g_free (unescaped);
-  g_clear_error (&error);
-  g_free (path);
-  if (file)
-    g_mapped_file_unref (file);
-
-  if (output)
-    g_list_free_full (output, (GDestroyNotify)g_bytes_unref);
+  g_list_free_full (output, (GDestroyNotify)g_bytes_unref);
 }
 
 /**
@@ -1442,7 +1423,7 @@ cockpit_web_response_file (CockpitWebResponse *response,
                            const gchar *escaped,
                            const gchar **roots)
 {
-  web_response_file (response, escaped, roots, NULL, NULL);
+  web_response_file (response, escaped, roots, FALSE, FALSE, NULL, NULL);
 }
 
 void
@@ -1451,7 +1432,16 @@ cockpit_web_response_template (CockpitWebResponse *response,
                                    const gchar **roots,
                                    GHashTable *values)
 {
-  web_response_file (response, escaped, roots, substitute_hash_value, values);
+  web_response_file (response, escaped, roots, FALSE, FALSE, substitute_hash_value, values);
+}
+
+void
+cockpit_web_response_file_or_gz (CockpitWebResponse *response,
+                                 gboolean accepts_gzip,
+                                 const gchar *escaped,
+                                 const gchar **roots)
+{
+  web_response_file (response, escaped, roots, TRUE, accepts_gzip, NULL, NULL);
 }
 
 static gboolean
@@ -1601,10 +1591,12 @@ load_file (const gchar *filename,
            GError **error)
 {
   GError *local_error = NULL;
-  GMappedFile *mapped;
-  GBytes *bytes;
 
-  mapped = g_mapped_file_new (filename, FALSE, &local_error);
+  g_autoptr(GMappedFile) mapped = g_mapped_file_new (filename, FALSE, &local_error);
+
+  if (mapped)
+    /* success! */
+    return g_mapped_file_get_bytes (mapped);
 
   if (g_error_matches (local_error, G_FILE_ERROR, G_FILE_ERROR_NOENT) ||
       g_error_matches (local_error, G_FILE_ERROR, G_FILE_ERROR_ISDIR) ||
@@ -1613,27 +1605,24 @@ load_file (const gchar *filename,
       g_error_matches (local_error, G_FILE_ERROR, G_FILE_ERROR_INVAL))
     {
       g_clear_error (&local_error);
-      return NULL;
     }
 
   /* A real error to stop on */
-  else if (local_error)
+  else
     {
       g_propagate_error (error, local_error);
-      return NULL;
     }
 
-  bytes = g_mapped_file_get_bytes (mapped);
-  g_mapped_file_unref (mapped);
-
-  return bytes;
+  return NULL;
 }
 
 /**
  * cockpit_web_response_negotiation:
  * @path: likely filesystem path
  * @existing: a table of existing files
- * @chosen: out, a pointer to the suffix that was chosen
+ * @language: requested client language
+ * @out_is_language_specific: a pointer to a gboolean whether the actual file is specific to @language
+ * @out_is_compressed: a pointer to a gboolean whether the actual file is compressed
  * @error: a failure
  *
  * Find a file to serve based on the suffixes. We prune off extra
@@ -1648,7 +1637,8 @@ GBytes *
 cockpit_web_response_negotiation (const gchar *path,
                                   GHashTable *existing,
                                   const gchar *language,
-                                  gchar **actual,
+                                  gboolean *out_is_language_specific,
+                                  gboolean *out_is_compressed,
                                   GError **error)
 {
   gchar *base = NULL;
@@ -1661,6 +1651,7 @@ cockpit_web_response_negotiation (const gchar *path,
   gchar *shorter = NULL;
   gchar *lang = NULL;
   gchar *lang_region = NULL;
+  gboolean is_language_specific, is_compressed;
 
   gint i;
 
@@ -1718,27 +1709,43 @@ cockpit_web_response_negotiation (const gchar *path,
             {
             case 0:
               name = g_strconcat (base, ".", lang_region, ext, NULL);
+              is_language_specific = TRUE;
+              is_compressed = FALSE;
               break;
             case 1:
               name = g_strconcat (base, ".", lang_region, ext, ".gz", NULL);
+              is_language_specific = TRUE;
+              is_compressed = TRUE;
               break;
             case 2:
               name = g_strconcat (base, ".", lang, ext, NULL);
+              is_language_specific = TRUE;
+              is_compressed = FALSE;
               break;
             case 3:
               name = g_strconcat (base, ".", lang, ext, ".gz", NULL);
+              is_language_specific = TRUE;
+              is_compressed = TRUE;
               break;
             case 4:
               name = g_strconcat (base, ext, NULL);
+              is_language_specific = FALSE;
+              is_compressed = FALSE;
               break;
             case 5:
               name = g_strconcat (base, ".min", ext, NULL);
+              is_language_specific = FALSE;
+              is_compressed = FALSE;
               break;
             case 6:
               name = g_strconcat (base, ext, ".gz", NULL);
+              is_language_specific = FALSE;
+              is_compressed = TRUE;
               break;
             case 7:
               name = g_strconcat (base, ".min", ext, ".gz", NULL);
+              is_language_specific = FALSE;
+              is_compressed = TRUE;
               break;
             default:
               g_assert_not_reached ();
@@ -1768,10 +1775,12 @@ cockpit_web_response_negotiation (const gchar *path,
 out:
   if (local_error)
     g_propagate_error (error, local_error);
-  if (bytes && name && actual)
+  if (bytes)
     {
-      *actual = name;
-      name = NULL;
+      if (out_is_language_specific)
+        *out_is_language_specific = is_language_specific;
+      if (out_is_compressed)
+        *out_is_compressed = is_compressed;
     }
   g_free (name);
   g_free (base);
@@ -1800,6 +1809,7 @@ cockpit_web_response_content_type (const gchar *path)
     { ".svg", "image/svg+xml" },
     { ".ttf", "application/octet-stream" }, /* unassigned */
     { ".txt", "text/plain" },
+    { ".wasm", "application/wasm" },
     { ".woff", "application/font-woff" },
     { ".xml", "text/xml" },
   };
@@ -1935,10 +1945,9 @@ cockpit_web_response_get_origin (CockpitWebResponse *self)
 }
 
 const gchar *
-cockpit_web_response_get_protocol (CockpitWebResponse *self,
-                                   GHashTable *headers)
+cockpit_web_response_get_protocol (CockpitWebResponse *self)
 {
-  return cockpit_connection_get_protocol (self->io, headers, self->flags & COCKPIT_WEB_RESPONSE_FOR_TLS_PROXY);
+  return self->protocol;
 }
 
 static void
@@ -1946,25 +1955,3 @@ cockpit_web_response_flow_iface_init (CockpitFlowInterface *iface)
 {
   /* No implementation */
 }
-
-const gchar *
-cockpit_connection_get_protocol (GIOStream *connection,
-                                 GHashTable *headers,
-                                 gboolean for_tls_proxy)
-{
-  const gchar *protocol = NULL;
-  const gchar *protocol_header;
-
-  if (connection && G_IS_TLS_CONNECTION (connection))
-    {
-      protocol = "https";
-    }
-  else
-    {
-      protocol_header = cockpit_conf_string ("WebService", "ProtocolHeader");
-      if (protocol_header && headers)
-         protocol = g_hash_table_lookup (headers, protocol_header);
-    }
-
-  return protocol ?: (for_tls_proxy ? "https" : "http");
-}
