@@ -23,9 +23,7 @@
 #include "cockpitwebresponse.h"
 #include "cockpitwebserver.h"
 
-#include "mock-io-stream.h"
-
-#include "common/cockpittest.h"
+#include "testlib/cockpittest.h"
 
 #include "websocket/websocket.h"
 
@@ -35,7 +33,7 @@
 #include <string.h>
 
 /* headers that are present in every request */
-#define STATIC_HEADERS "X-DNS-Prefetch-Control: off\r\nReferrer-Policy: no-referrer\r\nX-Content-Type-Options: nosniff\r\nCross-Origin-Resource-Policy: same-origin\r\n\r\n"
+#define STATIC_HEADERS "X-DNS-Prefetch-Control: off\r\nReferrer-Policy: no-referrer\r\nX-Content-Type-Options: nosniff\r\nCross-Origin-Resource-Policy: same-origin\r\nX-Frame-Options: sameorigin\r\n\r\n"
 static gchar *srcdir;
 
 typedef struct {
@@ -50,6 +48,7 @@ typedef struct {
     const gchar *path;
     const gchar *header;
     const gchar *value;
+    const gchar *expected_content_type;
     CockpitCacheType cache;
     gboolean for_tls_proxy;
 } TestFixture;
@@ -80,7 +79,7 @@ setup (TestCase *tc,
 
   input = g_memory_input_stream_new ();
   tc->output = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
-  io = mock_io_stream_new (input, tc->output);
+  io = g_simple_io_stream_new (input, tc->output);
   g_object_unref (input);
 
   if (fixture && fixture->header)
@@ -89,8 +88,8 @@ setup (TestCase *tc,
       g_hash_table_insert (headers, g_strdup (fixture->header), g_strdup (fixture->value));
     }
 
-  tc->response = cockpit_web_response_new (io, path, path, NULL, headers,
-                                           (fixture && fixture->for_tls_proxy) ? COCKPIT_WEB_RESPONSE_FOR_TLS_PROXY : COCKPIT_WEB_RESPONSE_NONE);
+  tc->response = cockpit_web_response_new (io, path, path, headers, "GET",
+                                           (fixture && fixture->for_tls_proxy) ? "https" : "http");
 
   if (headers)
     g_hash_table_unref (headers);
@@ -233,23 +232,12 @@ test_return_gerror_headers (TestCase *tc,
     "\r\n");
 }
 
-static void
-test_return_error_resource (TestCase *tc,
-                            gconstpointer user_data)
-{
-  const gchar *roots[] = { srcdir, NULL };
-  cockpit_web_failure_resource = "/org/cockpit-project/Cockpit/fail.html";
-  cockpit_web_response_file (tc->response, "/non-existant", roots);
-  cockpit_assert_strmatch (output_as_string (tc), "HTTP/1.1 404 Not Found*<img*Not Found*");
-  cockpit_web_failure_resource = NULL;
-}
-
 static void
 test_file_not_found (TestCase *tc,
                      gconstpointer user_data)
 {
   const gchar *roots[] = { srcdir, NULL };
-  cockpit_web_response_file (tc->response, "/non-existant", roots);
+  cockpit_web_response_file (tc->response, "/non-existent", roots);
   cockpit_assert_strmatch (output_as_string (tc), "HTTP/1.1 404 Not Found*");
 }
 
@@ -332,7 +320,7 @@ test_file_breakout_non_existant (TestCase *tc,
 {
   gchar *root = realpath ( SRCDIR "/src", NULL);
   const gchar *roots[] = { root, NULL };
-  const gchar *breakout = "/../non-existant";
+  const gchar *breakout = "/../non-existent";
   gchar *check = g_build_filename (roots[0], breakout, NULL);
   g_assert (root);
   g_assert (!g_file_test (check, G_FILE_TEST_EXISTS));
@@ -342,14 +330,26 @@ test_file_breakout_non_existant (TestCase *tc,
   free (root);
 }
 
-static const TestFixture content_type_fixture = {
-  .path = "/pkg/shell/index.html"
+static const TestFixture content_type_fixture_html = {
+  .path = "/pkg/shell/index.html",
+  .expected_content_type = "text/html",
+};
+
+static const TestFixture content_type_fixture_png = {
+  .path = "/pkg/shell/images/server-small.png",
+  .expected_content_type = "image/png",
+};
+
+static const TestFixture content_type_fixture_wasm = {
+  .path = "/src/common/mock-content/test.wasm",
+  .expected_content_type = "application/wasm",
 };
 
 static void
 test_content_type (TestCase *tc,
                    gconstpointer user_data)
 {
+  const TestFixture *fixture = user_data;
   const gchar *roots[] = { srcdir, NULL };
   GHashTable *headers;
   const gchar *resp;
@@ -357,8 +357,6 @@ test_content_type (TestCase *tc,
   guint status;
   gssize off;
 
-  g_assert (user_data == &content_type_fixture);
-
   cockpit_web_response_file (tc->response, NULL, roots);
 
   resp = output_as_string (tc);
@@ -371,7 +369,7 @@ test_content_type (TestCase *tc,
   off = web_socket_util_parse_headers (resp + off, length - off, &headers);
   g_assert_cmpuint (off, >, 0);
 
-  g_assert_cmpstr (g_hash_table_lookup (headers, "Content-Type"), ==, "text/html");
+  g_assert_cmpstr (g_hash_table_lookup (headers, "Content-Type"), ==, fixture->expected_content_type);
 
   g_hash_table_unref (headers);
 }
@@ -397,19 +395,14 @@ test_template (TestCase *tc,
   g_hash_table_unref (data);
 }
 
-static const TestFixture cache_forever_fixture = {
-  .path = "/pkg/shell/index.html",
-  .cache = COCKPIT_WEB_RESPONSE_CACHE_FOREVER,
-};
-
 static const TestFixture cache_none_fixture = {
   .path = "/pkg/shell/index.html",
   .cache = COCKPIT_WEB_RESPONSE_NO_CACHE
 };
 
-static const TestFixture cache_private_fixture = {
+static const TestFixture cache_fixture = {
   .path = "/pkg/shell/index.html",
-  .cache = COCKPIT_WEB_RESPONSE_CACHE_PRIVATE
+  .cache = COCKPIT_WEB_RESPONSE_CACHE
 };
 
 static const TestFixture cache_unset_fixture = {
@@ -442,16 +435,14 @@ test_cache (TestCase *tc,
   off = web_socket_util_parse_headers (resp + off, length - off, &headers);
   g_assert_cmpuint (off, >, 0);
 
-  if (fixture->cache == COCKPIT_WEB_RESPONSE_CACHE_PRIVATE)
+  if (fixture->cache == COCKPIT_WEB_RESPONSE_CACHE)
     g_assert_cmpstr (g_hash_table_lookup (headers, "Vary"), ==, "Cookie");
   else
     g_assert_null (g_hash_table_lookup (headers, "Vary"));
 
-  if (fixture->cache == COCKPIT_WEB_RESPONSE_CACHE_FOREVER)
-    g_assert_cmpstr (g_hash_table_lookup (headers, "Cache-Control"), ==, "max-age=31556926, public");
-  else if (fixture->cache == COCKPIT_WEB_RESPONSE_NO_CACHE)
+  if (fixture->cache == COCKPIT_WEB_RESPONSE_NO_CACHE)
     g_assert_cmpstr (g_hash_table_lookup (headers, "Cache-Control"), ==, "no-cache, no-store");
-  else if (fixture->cache == COCKPIT_WEB_RESPONSE_CACHE_PRIVATE)
+  else if (fixture->cache == COCKPIT_WEB_RESPONSE_CACHE)
     g_assert_cmpstr (g_hash_table_lookup (headers, "Cache-Control"), ==, "max-age=86400, private");
   else
     g_assert_null (g_hash_table_lookup (headers, "Cache-Control"));
@@ -1087,7 +1078,7 @@ setup_plain (TestPlain *tc,
   input = g_memory_input_stream_new ();
   output = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
 
-  tc->io = mock_io_stream_new (input, output);
+  tc->io = g_simple_io_stream_new (input, output);
 
   g_object_unref (input);
   g_object_unref (output);
@@ -1111,7 +1102,7 @@ test_pop_path (TestPlain *tc,
   gchar *part;
   const gchar *start = "/cockpit/@localhost/another/test.html";
 
-  response = cockpit_web_response_new (tc->io, start, start, NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, start, start, tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, start);
   g_assert_cmpstr (cockpit_web_response_get_url_root (response), ==, NULL);
 
@@ -1151,7 +1142,7 @@ test_pop_path_root (TestPlain *tc,
   CockpitWebResponse *response;
   gchar *part;
 
-  response = cockpit_web_response_new (tc->io, "/", "/", NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, "/", "/", tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/");
 
   part = cockpit_web_response_pop_path (response);
@@ -1170,7 +1161,7 @@ test_skip_path (TestPlain *tc,
   CockpitWebResponse *response;
   const gchar *start = "/cockpit/@localhost/another/test.html";
 
-  response = cockpit_web_response_new (tc->io, start, start, NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, start, start, tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/cockpit/@localhost/another/test.html");
 
   g_assert (cockpit_web_response_skip_path (response) == TRUE);
@@ -1198,7 +1189,7 @@ test_skip_path_root (TestPlain *tc,
 {
   CockpitWebResponse *response;
 
-  response = cockpit_web_response_new (tc->io, "/", "/", NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, "/", "/", tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/");
 
   g_assert (cockpit_web_response_skip_path (response) == FALSE);
@@ -1214,31 +1205,31 @@ test_removed_prefix (TestPlain *tc,
 {
   CockpitWebResponse *response;
 
-  response = cockpit_web_response_new (tc->io, "/", "/", NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, "/", "/", tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/");
   g_assert_cmpstr (cockpit_web_response_get_url_root (response), ==, NULL);
   cockpit_web_response_abort (response);
   g_clear_object (&response);
 
-  response = cockpit_web_response_new (tc->io, "/path/", "/path/", NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, "/path/", "/path/", tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/path/");
   g_assert_cmpstr (cockpit_web_response_get_url_root (response), ==, NULL);
   cockpit_web_response_abort (response);
   g_clear_object (&response);
 
-  response = cockpit_web_response_new (tc->io, "/path/path2/", "/path2/", NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, "/path/path2/", "/path2/", tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/path2/");
   g_assert_cmpstr (cockpit_web_response_get_url_root (response), ==, "/path");
   cockpit_web_response_abort (response);
   g_clear_object (&response);
 
-  response = cockpit_web_response_new (tc->io, "/mis/", "/match/", NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, "/mis/", "/match/", tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, "/match/");
   g_assert_cmpstr (cockpit_web_response_get_url_root (response), ==, NULL);
   cockpit_web_response_abort (response);
   g_clear_object (&response);
 
-  response = cockpit_web_response_new (tc->io, NULL, NULL, NULL, tc->headers, COCKPIT_WEB_RESPONSE_NONE);
+  response = cockpit_web_response_new (tc->io, NULL, NULL, tc->headers, "GET", NULL);
   g_assert_cmpstr (cockpit_web_response_get_path (response), ==, NULL);
   g_assert_cmpstr (cockpit_web_response_get_url_root (response), ==, NULL);
   cockpit_web_response_abort (response);
@@ -1313,35 +1304,35 @@ test_gunzip_invalid (void)
 static void
 test_negotiation_first (void)
 {
-  gchar *chosen = NULL;
+  gboolean is_language_specific, is_compressed;
   GError *error = NULL;
   GBytes *bytes;
 
   bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/test-file.txt",
-                                            NULL, NULL, &chosen, &error);
+                                            NULL, NULL, &is_language_specific, &is_compressed, &error);
 
   cockpit_assert_bytes_eq (bytes, "A small test file\n", -1);
   g_assert_no_error (error);
   g_bytes_unref (bytes);
 
-  g_assert_cmpstr (chosen, ==, SRCDIR "/src/common/mock-content/test-file.txt");
-  g_free (chosen);
+  g_assert (!is_language_specific);
+  g_assert (!is_compressed);
 }
 
 static void
 test_negotiation_last (void)
 {
-  gchar *chosen = NULL;
+  gboolean is_language_specific, is_compressed;
   GError *error = NULL;
   gchar *checksum;
   GBytes *bytes;
 
   bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/large.js",
-                                            NULL, NULL, &chosen, &error);
+                                            NULL, NULL, &is_language_specific, &is_compressed, &error);
 
   g_assert_no_error (error);
-  g_assert_cmpstr (chosen, ==, SRCDIR "/src/common/mock-content/large.min.js.gz");
-  g_free (chosen);
+  g_assert (!is_language_specific);
+  g_assert (is_compressed);
 
   checksum = g_compute_checksum_for_bytes (G_CHECKSUM_MD5, bytes);
   g_assert_cmpstr (checksum, ==, "e5284b625b7665fc04e082827de3436c");
@@ -1353,19 +1344,15 @@ test_negotiation_last (void)
 static void
 test_negotiation_prune (void)
 {
-  gchar *chosen = NULL;
   GError *error = NULL;
   GBytes *bytes;
 
   bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/test-file.extra.extension.txt",
-                                            NULL, NULL, &chosen, &error);
+                                            NULL, NULL, NULL, NULL, &error);
 
   cockpit_assert_bytes_eq (bytes, "A small test file\n", -1);
   g_assert_no_error (error);
   g_bytes_unref (bytes);
-
-  g_assert_cmpstr (chosen, ==, SRCDIR "/src/common/mock-content/test-file.txt");
-  g_free (chosen);
 }
 
 static void
@@ -1380,7 +1367,7 @@ test_negotiation_with_listing (void)
   g_hash_table_add (existing, SRCDIR "/src/common/mock-content/test-file.txt.gz");
 
   bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/test-file.txt",
-                                            existing, NULL, NULL, &error);
+                                            existing, NULL, NULL, NULL, &error);
 
   cockpit_assert_bytes_eq (bytes, "\x1F\x8B\x08\x08N1\x03U\x00\x03test-file.txt\x00"
                            "sT(\xCEM\xCC\xC9Q(I-.QH\xCB\xCCI\xE5\x02\x00>PjG\x12\x00\x00\x00", 52);
@@ -1393,59 +1380,57 @@ test_negotiation_with_listing (void)
 static void
 test_negotiation_locale (void)
 {
-  gchar *chosen = NULL;
+  gboolean is_language_specific, is_compressed;
   GError *error = NULL;
   GBytes *bytes;
 
   bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/test-file.txt",
-                                            NULL, "zh-cn", &chosen, &error);
+                                            NULL, "zh-cn", &is_language_specific, &is_compressed, &error);
 
   cockpit_assert_bytes_eq (bytes, "A translated test file\n", -1);
   g_assert_no_error (error);
   g_bytes_unref (bytes);
 
-  g_assert_cmpstr (chosen, ==, SRCDIR "/src/common/mock-content/test-file.zh_CN.txt");
-  g_free (chosen);
+  g_assert (is_language_specific);
+  g_assert (!is_compressed);
 }
 
 static void
 test_negotiation_notfound (void)
 {
-  gchar *chosen = NULL;
   GError *error = NULL;
   GBytes *bytes;
 
-  bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/non-existant",
-                                            NULL, NULL, &chosen, &error);
+  bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/non-existent",
+                                            NULL, NULL, NULL, NULL, &error);
 
   g_assert_no_error (error);
   g_assert (bytes == NULL);
-
-  g_assert (chosen == NULL);
 }
 
 static void
 test_negotiation_failure (void)
 {
-  gchar *chosen = NULL;
   GError *error = NULL;
   GBytes *bytes;
 
   bytes = cockpit_web_response_negotiation (SRCDIR "/src/common/mock-content/directory",
-                                            NULL, NULL, &chosen, &error);
+                                            NULL, NULL, NULL, NULL, &error);
 
   g_assert (error != NULL);
   g_error_free (error);
 
   g_assert (bytes == NULL);
-
-  g_assert (chosen == NULL);
 }
 
 int
 main (int argc,
       char *argv[])
 {
+  extern const gchar *cockpit_webresponse_fail_html_text;
+  cockpit_webresponse_fail_html_text =
+    "<html><head><title>@@message@@</title></head><body>@@message@@</body></html>\n";
+
   gint ret;
 
   srcdir = realpath (SRCDIR, NULL);
@@ -1463,8 +1448,6 @@ main (int argc,
               setup, test_return_error_headers, teardown);
   g_test_add ("/web-response/return-gerror-headers", TestCase, NULL,
               setup, test_return_gerror_headers, teardown);
-  g_test_add ("/web-response/return-error-resource", TestCase, NULL,
-              setup, test_return_error_resource, teardown);
   g_test_add ("/web-response/file/not-found", TestCase, NULL,
               setup, test_file_not_found, teardown);
   g_test_add ("/web-response/file/directory-denied", TestCase, NULL,
@@ -1481,7 +1464,11 @@ main (int argc,
               setup, test_file_breakout_non_existant, teardown);
   g_test_add ("/web-reponse/file/template", TestCase, &template_fixture,
               setup, test_template, teardown);
-  g_test_add ("/web-response/content-type", TestCase, &content_type_fixture,
+  g_test_add ("/web-response/content-type/html", TestCase, &content_type_fixture_html,
+              setup, test_content_type, teardown);
+  g_test_add ("/web-response/content-type/png", TestCase, &content_type_fixture_png,
+              setup, test_content_type, teardown);
+  g_test_add ("/web-response/content-type/wasm", TestCase, &content_type_fixture_wasm,
               setup, test_content_type, teardown);
   g_test_add ("/web-response/content-encoding", TestCase, NULL,
               setup, test_content_encoding, teardown);
@@ -1504,9 +1491,7 @@ main (int argc,
   g_test_add ("/web-response/origin-tls-proxy", TestCase, &fixture_origin_tls_proxy,
               setup, test_origin, teardown);
 
-  g_test_add ("/web-response/cache-forever", TestCase, &cache_forever_fixture,
-              setup, test_cache, teardown);
-  g_test_add ("/web-response/cache-private", TestCase, &cache_private_fixture,
+  g_test_add ("/web-response/cache", TestCase, &cache_fixture,
               setup, test_cache, teardown);
   g_test_add ("/web-response/cache-none", TestCase, &cache_none_fixture,
               setup, test_cache, teardown);
